﻿using System;
using System.Threading.Tasks;
using Windows.Devices.Custom;
using System.Runtime.InteropServices.WindowsRuntime;
using Windows.Storage.Streams;
using Windows.Foundation;
using System.Collections.Generic;
using System.Diagnostics;
using System.Reflection;
using Windows.Devices.Enumeration;
using Windows.Devices.Usb;
using FTDI.D2XX.Device;
using System.Linq;

namespace DeviceWatcher
{

    /// <summary>
    /// Interface providing details of an FTDI device connected to the system.
    /// </summary>
    public class FTDeviceInfoNode //: INotifyPropertyChanged
    {
        /// <summary>
        /// Enum value representing the device type.
        /// </summary>
        DEVICE_TYPE DeviceType { get; }

        /// <summary>
        /// USB Manufacturer Device Descriptor string.
        /// </summary>
        String Manufacturer { get; }

        /// <summary>
        /// USB Device Descriptor string.
        /// </summary>
        public String Description { get; }

        /// <summary>
        /// Device serial number string.
        /// </summary>
        public String SerialNumber { get; }

        /// <summary>
        /// Unique string that identifies an FTDI device.
        /// </summary>
        public String DeviceId;

        /// <summary>
        /// Returns the open state of the current device.
        /// </summary>
        Boolean IsOpen { get; }
    }


    /// <summary>
    /// Management class for connected FTDI devices.  Allows you to enumerate and communicate with connected 
    /// FTDI devices.
    /// </summary>
    public sealed class FTManager
    {
        #region CONSTANTS
        /// <summary>
        /// Enumeration of FTDI Product IDs.
        /// </summary>
        private enum FTDI_DEVICE_PIDS : uint
        {
            FT_PID_6001 = 0x6001,
            FT_PID_6010 = 0x6010,
            FT_PID_6011 = 0x6011,
            FT_PID_6014 = 0x6014,
            FT_PID_6015 = 0x6015,
        }

        private enum FT_BCD_REVISION : uint
        {
            FT232AM = 0x0200,
            FT232BM = 0x0400,
            FT2232D = 0x0500,
            FT232R = 0x0600,
            FT2232H = 0x0700,
            FT4232H = 0x0800,
            FT232H = 0x0900,
            FTXSeries = 0x1000,
        }

        /// <summary>
        /// Constant value of FTDI Vendor ID.
        /// </summary>
        private const uint FTDI_VID = 0x0403;
        #endregion

        // Internal device list.
        private static readonly List<FTD2xxDevice> FtdiDeviceList = new List<FTD2xxDevice>();
        private static readonly object syncLock = new object();
        private static readonly List<VendorIdProductId> listDeviceWatchers = new List<VendorIdProductId>
        {
            new VendorIdProductId(FTDI_VID, 0x6001),
        };

        private class VendorIdProductId : IEquatable<VendorIdProductId>
        {
            public VendorIdProductId(uint vid, uint pid)
            {
                VendorId = vid;
                ProductId = pid;
                FtWatcher = null;
            }

            public uint VendorId { get; private set; }

            public uint ProductId { get; private set; }

            public Windows.Devices.Enumeration.DeviceWatcher FtWatcher { get; set; }

            public bool Equals(VendorIdProductId other)
            {
                return other.VendorId == VendorId && other.ProductId == ProductId;
            }
        }

        /// <summary>
        /// Default constructor for FTManager class.
        /// </summary>
        public FTManager()
        {
            // Create watchers for each item in the VID PID list.
            foreach (var i in listDeviceWatchers.Where(i => i.FtWatcher == null))
            {
                GenerateFtWatcher(i);
                StartWatcher(i.FtWatcher);
            }
        }

        private static void GenerateFtWatcher(VendorIdProductId vidpid)
        {
            string ftSelector = "System.Devices.InterfaceClassGuid:=\"{219D0508-57A8-4FF5-97A1-BD86587C6C7E}\"";

            vidpid.FtWatcher = DeviceInformation.CreateWatcher(ftSelector);
        }

        /// <summary>
        /// Create a watcher for the specified VID and PID.
        /// </summary>
        /// <param name="ftWatcher">Device watcher.</param>
        private void StartWatcher(Windows.Devices.Enumeration.DeviceWatcher ftWatcher)
        {
            if (ftWatcher == null)
                return;

            ftWatcher.Added   += OnDeviceAdded;
            ftWatcher.Removed += OnDeviceRemoved;
            ftWatcher.Updated += ftWatcher_Updated;
            ftWatcher.Stopped += ftWatcher_Stopped;

            // Start the watcher...
            ftWatcher.Start();
        }

        void ftWatcher_Stopped(DeviceWatcher sender, object args)
        {
            //throw new NotImplementedException();

        }

        void ftWatcher_Updated(DeviceWatcher sender, DeviceInformationUpdate args)
        {
            // throw new NotImplementedException();
        }

        private void StopWatcher(DeviceWatcher ftWatcher)
        {
            ftWatcher.Stop();
            ftWatcher.Added   -= OnDeviceAdded;
            ftWatcher.Removed -= OnDeviceRemoved;
            ftWatcher.Updated -= ftWatcher_Updated;
            ftWatcher.Stopped -= ftWatcher_Stopped;
        }

        /// <summary>
        /// OnDeviceRemoved callback.
        /// </summary>
        private void OnDeviceRemoved(DeviceWatcher sender, DeviceInformationUpdate deviceInformationUpdate)
        {
            DeviceRemoved(deviceInformationUpdate);
        }

        /// <summary>
        /// Remove device from list of devices.
        /// </summary>
        private void DeviceRemoved(DeviceInformationUpdate deviceInformationUpdate)
        {
            var ftDevice = FindDeviceInList(deviceInformationUpdate.Id);

            if (ftDevice == null)
                return;

            if (ftDevice.IsOpen)
                ftDevice.Close();
            RemoveDeviceFromList(ftDevice);
        }

        /// <summary>
        /// The application should call this when application execution is being suspended.
        /// This will cause the appropriate driver objects to be disposed.
        /// </summary>
        
        public void Suspend()
        {
            while (FtdiDeviceList != null && FtdiDeviceList.Count > 0)
            {

                var ftDevice = FtdiDeviceList[0];
                {
                    if (ftDevice == null)
                        continue;

                    if (ftDevice.IsOpen)
                        ftDevice.Close();                    
                    //((FTDevice)ftDevice).Dispose();
                    RemoveDeviceFromList(ftDevice);
                }
            }

            if (listDeviceWatchers != null)
            {
                for (int c = 0; c < listDeviceWatchers.Count; c++ )
                {
                    if (listDeviceWatchers[c] != null && listDeviceWatchers[c].FtWatcher != null)
                    {
                        if (listDeviceWatchers[c].FtWatcher.Status == DeviceWatcherStatus.Started ||
                           listDeviceWatchers[c].FtWatcher.Status == DeviceWatcherStatus.EnumerationCompleted)
                        {
                            StopWatcher(listDeviceWatchers[c].FtWatcher);
                        }
                    }
                }
            }
        }

        /// <summary>
        /// This function must be called by the application on resumption following a suspend,
        /// It is responsible for re starting the appropriate list of USB Device watchers.
        /// </summary>
        public void Resume()
        {          
            if (listDeviceWatchers != null)
            {
                for (int i = 0; i < listDeviceWatchers.Count; i++)               
                {
                    if (listDeviceWatchers[i] != null && 
                        listDeviceWatchers[i].FtWatcher != null 
                        && listDeviceWatchers[i].FtWatcher.Status != DeviceWatcherStatus.Started &&
                        listDeviceWatchers[i].FtWatcher.Status != DeviceWatcherStatus.EnumerationCompleted
                        )
                    {                      
                        StartWatcher(listDeviceWatchers[i].FtWatcher);
                    }
                }
            }
        }


        /// <summary>
        /// OnDeviceAdded callback.
        /// </summary>
        private async void OnDeviceAdded(DeviceWatcher sender, DeviceInformation deviceInformation)
        {
            await Task.Run(async () =>
            {
                await DeviceAdded(deviceInformation);
            });
        }

        /// <summary>
        /// Called when a new FTDI device is added to the system.
        /// </summary>
        private async Task<Boolean> DeviceAdded(DeviceInformation deviceInformation)
        {
            try
            {
                if (FindDeviceInList(deviceInformation.Id) != null) // Device is already in our list, return.
                    return false;

                Task<CustomDevice> t = null;
                await Windows.ApplicationModel.Core.CoreApplication.MainView.CoreWindow.Dispatcher.RunAsync(Windows.UI.Core.CoreDispatcherPriority.High,
                () =>
                {
                    t = CustomDevice.FromIdAsync(deviceInformation.Id, DeviceAccessMode.ReadWrite, DeviceSharingMode.Exclusive).AsTask();
                });

                bool timeout = t.Wait(10000); // wait 10 secs for the user to respond, otherwise just return.
                if (!timeout)
                    return false;

                var device = t.Result;
                if (device == null)
                    return false;

                //var deviceAccessStatus = DeviceAccessInformation.CreateFromId(deviceInformation.Id).CurrentStatus;

                FTDeviceInfoNode node = new FTDeviceInfoNode
                {
                    DeviceId = deviceInformation.Id
                };

                FTD2xxDevice dev = new FTD2xxDevice(node,device);
                AddDeviceToList(dev);

            }
            catch (Exception ex)
            {
                Debug.WriteLine(ex.Message);
                return false;
            }
            return true;
        }

        /// <summary>
        /// Returns a list of FTDI devices connected to the system.
        /// </summary>
        /// <returns>IList containing FTDeviceInfoNodes.</returns>
        /// <example> This example shows how to access an FTDI device using the FTManager class.
        /// <code>
        ///    FTManager ftManager = new FTManager();
        ///    var deviceList = ftManager.GetDeviceList();
        ///
        ///    if (deviceList.Count > 0)
        ///    {
        ///        byte[] data = new byte[10];
        ///        // Open a handle to the device.
        ///        IFTDevice ftDevice = ftManager.OpenByDeviceID(deviceList[0].DeviceId);
        ///
        ///        await ftDevice.SetBaudRate(9600);
        ///        await ftDevice.SetFlowControl(FLOW_CONTROL.RTS_CTS, 0, 0);
        ///        // Write/Read some data.
        ///        await ftDevice.Write(data, 10);
        ///        ftDevice.Read(data, 10);
        ///        // Close the device handle.
        ///        ftDevice.Close();
        ///    }
        /// </code>
        /// </example>
        public IList<FTDeviceInfoNode> GetDeviceList()
        {
            lock (syncLock)
            {
                return CreateDeviceInfoList();
            }
        }

        // Populate the device info list.
        private IList<FTDeviceInfoNode> CreateDeviceInfoList()
        {
            List<FTDeviceInfoNode> deviceInfoList = new List<FTDeviceInfoNode>();

            foreach (FTD2xxDevice ftDevice in FtdiDeviceList)
            {
                deviceInfoList.Add(ftDevice.DeviceInfoNode);
            }

            return deviceInfoList;
        }

        private void AddDeviceToList(FTD2xxDevice ftDevice)
        {
            lock (syncLock)
            {
                FtdiDeviceList.Add(ftDevice);
            }
        }

        private void RemoveDeviceFromList(FTD2xxDevice ftDevice)
        {
            if (ftDevice == null)
                return;

            lock (syncLock)
            {
                FtdiDeviceList.Remove(ftDevice);
            }
        }

        // Try and find this device in our device list.
        private FTD2xxDevice FindDeviceInList(String id)
        {
            lock (syncLock)
            {
                foreach (var dev in FtdiDeviceList)
                {
                    if (dev.DeviceInfoNode.DeviceId == id)
                        return dev;
                }
            }
            return null;
        }

        /// <summary>
        /// Add a custom vendor ID/product ID combination to the list of allowed devices.
        /// </summary>
        /// <param name="vendorId">Device Vendor ID.</param>
        /// <param name="productId">Device Product ID.</param>
        /// <returns>True on success, false if the vendor ID/product ID combination is already in the list.</returns>
        /// <example> This example shows how to add a custom VID and PID to the driver list.
        /// <code>
        ///    uint myVid = 0x1234;
        ///    uint myPid = 0x4321;
        ///    
        ///    FTManager.AddVIDPID(myVid, myPid);
        ///    FTManager ftManager = new FTManager();
        /// </code>
        /// </example>
        public static Boolean AddVIDPID(uint vendorId, uint productId)
        {
            if (vendorId == 0x0000 || productId == 0x0000)
                return false;

            try
            {
                int index = FindVIDPID(vendorId, productId);
                if (index > -1)
                { // already in the list.
                    return false;
                }

                VendorIdProductId newItem = new VendorIdProductId(vendorId, productId);
                listDeviceWatchers.Add(newItem);
                return true;
            }
            catch
            {
                Debug.WriteLine("Failed to add item.");
                return false;
            }
        }

        private static int FindVIDPID(uint vendorId, uint productId)
        {
            VendorIdProductId findMe = new VendorIdProductId(vendorId, productId);
            for (int i = 0; i < listDeviceWatchers.Count; i++)
            {
                if (findMe.Equals(listDeviceWatchers[i]))
                    return i;
            }

            return -1;
        }

        /// <summary>
        /// Remove a custom vendor ID/product ID combination from the list of allowed devices.
        /// </summary>
        /// <param name="vendorId">Device Vendor ID.</param>
        /// <param name="productId">Device Product ID.</param>
        /// <returns>True on success, false if the vendor ID/product ID combination was not found in the list.</returns>
        /// <example> This example shows how to remove a standard FTDI VID and PID from the driver list.
        /// <code>
        ///    uint rDeviceVid = 0x0403;
        ///    uint rDevicePid = 0x6001;
        ///    
        ///    FTManager.RemoveVIDPID(rDeviceVid, rDevicePid);
        ///    FTManager ftManager = new FTManager();
        /// </code>
        /// </example>
        public static Boolean RemoveVIDPID(uint vendorId, uint productId)
        {
            try
            {
                int index = FindVIDPID(vendorId, productId);
                if (index > -1)
                {
                    return listDeviceWatchers.Remove(listDeviceWatchers[index]);
                }
                // not in the list.
                return false;
            }
            catch
            {
                Debug.WriteLine("Failed to remove item.");
                return false;
            }
        }

        /// <summary>
        /// Open a handle to the specified FTDI device.
        /// </summary>
        /// <param name="id">Unique ID of the deivce to be opened.</param>
        /// <returns>An open FTDI device on success, null otherwise.</returns>
        public FTD2xxDevice OpenByDeviceID(String id)
        {
            lock (syncLock)
            {
                foreach (FTD2xxDevice ftDevice in FtdiDeviceList)
                {
                    FTDeviceInfoNode node = ftDevice.DeviceInfoNode;
                    if (node.DeviceId.Equals(id))
                    {
                        var device = ftDevice; // as FTD2xxDevice; // FTDevice;
                        if (device != null)
                            return (FTD2xxDevice) FTD2xxDevice.FT_Open(id, 0);
                    }
                }
            }
            return null;
        }

        /// <summary>
        /// Returns the version string of this library.
        /// </summary>
        public static String GetLibraryVersion()
        {
            var asmName = typeof(FTManager).GetTypeInfo().AssemblyQualifiedName;
            var versionExpression = new System.Text.RegularExpressions.Regex("Version=(?<version>[0-9.]*)");
            var m = versionExpression.Match(asmName);
            string version = String.Empty;
            if (m.Success)
            {
                version = m.Groups["version"].Value;
            }

            return version;
        }
    }
}
